# (c) 2016 Anaconda, Inc. / https://anaconda.com
# All Rights Reserved
#
# constructor is distributed under the terms of the BSD 3-clause license.
# Consult LICENSE.txt or http://opensource.org/licenses/BSD-3-Clause.

import hashlib
import logging
import math
import re
import sys
from os import sep, unlink
from os.path import basename, isdir, isfile, islink, normpath
from shutil import rmtree

try:
    import yaml
except ImportError:
    import ruamel_yaml as yaml

logger = logging.getLogger(__name__)


def filename_dist(dist):
    """ Return the filename of a distribution. """
    if hasattr(dist, 'to_filename'):
        return dist.to_filename()
    else:
        return dist


def fill_template(data, d, exceptions=[]):
    pat = re.compile(r'__(\w+)__')

    def replace(match):
        key = match.group(1)
        return key if key in exceptions else d[key]

    return pat.sub(replace, data)


def hash_files(paths, algorithm='md5'):
    h = hashlib.new(algorithm)
    for path in paths:
        with open(path, 'rb') as fi:
            while True:
                chunk = fi.read(262144)
                if not chunk:
                    break
                h.update(chunk)
    return h.hexdigest()


def make_VIProductVersion(version):
    """
    always create a version of the form X.X.X.X
    """
    pat = re.compile(r'\d+$')
    res = []
    for part in version.split('.'):
        if pat.match(part):
            res.append(part)
    while len(res) < 4:
        res.append('0')
    return '.'.join(res[:4])


def read_ascii_only(path):
    with open(path) as fi:
        data = fi.read()
    for c in data:
        if ord(c) > 127:
            sys.exit("Error: unexpected non-ASCII character '%s' in: %s" %
                     (c, path))
    return data


if_pat = re.compile(r'^#if ([ \S]+)$\n'
                    r'(.*?)'
                    r'(^#else\s*$\n(.*?))?'
                    r'^#endif\s*$\n', re.M | re.S)


def preprocess(data, namespace):

    def if_repl(match):
        cond = match.group(1).strip()
        if eval(cond, namespace, {}):
            return match.group(2)
        else:
            return match.group(4) or ''

    return if_pat.sub(if_repl, data)


def add_condarc(info):
    condarc = info.get('condarc')
    if condarc is None:
        # The legacy approach
        write_condarc = info.get('write_condarc')
        default_channels = info.get('conda_default_channels')
        channel_alias = info.get('conda_channel_alias')
        channels = info.get('channels')
        if not (write_condarc and (default_channels or channels or channel_alias)):
            return
        condarc = {}
        if default_channels:
            condarc['default_channels'] = default_channels
        if channels:
            condarc['channels'] = channels
        if channel_alias:
            condarc['channel_alias'] = channel_alias
    if isinstance(condarc, dict):
        condarc = yaml.dump(condarc)
    yield '# ----- add condarc'
    if info['_platform'].startswith('win'):
        yield 'Var /Global CONDARC'
        yield 'FileOpen $CONDARC "$INSTDIR\\.condarc" w'
        for line in condarc.splitlines():
            yield 'FileWrite $CONDARC "%s$\\r$\\n"' % line
        yield 'FileClose $CONDARC'
    else:
        yield 'cat <<EOF >"$PREFIX/.condarc"'
        for line in condarc.splitlines():
            yield line
        yield 'EOF'


def ensure_transmuted_ext(info, url):
    """
    If transmuting, micromamba won't find the dist in the preconda tarball
    unless it has the (correct and transmuted) extension. Otherwise, the command
    `micromamba constructor --extract-tarballs` fails.
    Unfortunately this means the `urls` file might end up containing
    fake URLs, since those .conda archives might not really exist online,
    and they were only created locally.
    """
    if (
        info.get("transmute_file_type") == ".conda"
        and "micromamba" in basename(info.get("_conda_exe", ""))
    ):
        if url.lower().endswith(".tar.bz2"):
            url = url[:-8] + ".conda"
    return url


def get_final_url(info, url):
    mapping = info.get('channels_remap', [])
    for entry in mapping:
        src = entry['src']
        dst = entry['dest']
        if url.startswith(src):
            new_url = url.replace(src, dst)
            if url.endswith(".tar.bz2"):
                logger.warning("You need to make the package %s available "
                               "at %s", url.rsplit('/', 1)[1], new_url)
            return new_url
    return url


def get_final_channels(info):
    mapped_channels = []
    for channel in info.get('channels', []):
        url = get_final_url(info, channel)
        if url.startswith("file://"):
            logger.warning("local channel %s does not have a remap. "
                           "It will not be included in the installer", url)
            continue
        mapped_channels.append(url)
    return mapped_channels


def normalize_path(path):
    new_path = normpath(path)
    return new_path.replace(sep + sep, sep)


def rm_rf(path):
    """
    try to delete path, but never fail
    """
    try:
        if islink(path) or isfile(path):
            # Note that we have to check if the destination is a link because
            # exists('/path/to/dead-link') will return False, although
            # islink('/path/to/dead-link') is True.
            unlink(path)
        elif isdir(path):
            rmtree(path)
    except OSError:
        pass


def yield_lines(path):
    for line in open(path):
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        yield line


def approx_size_kb(info, which="pkgs"):
    valid = ("pkgs", "tarballs", "total")
    assert which in valid, f"'which' must be one of {valid}"
    size_pkgs = info.get('_approx_pkgs_size', 0)
    size_tarballs = info.get('_approx_tarballs_size', 0)
    if which == "pkgs":
        size_bytes = size_pkgs
    elif which == "tarballs":
        size_bytes = size_tarballs
    else:
        size_bytes = size_pkgs + size_tarballs

    # division by 10^3 instead of 2^10 is deliberate here. gives us more room
    return int(math.ceil(size_bytes/1000))



#!/usr/bin/env python
"""Run examples bundled with this repo."""
import argparse
import os
import platform
import shutil
import subprocess
import sys
import tempfile
import time
from datetime import timedelta
from pathlib import Path

from constructor.utils import rm_rf

try:
    import coverage  # noqa
    COV_CMD = ['coverage', 'run', '--branch', '--append', '-m']
except ImportError:
    COV_CMD = []


HERE = os.path.abspath(os.path.dirname(__file__))
REPO_DIR = os.path.dirname(HERE)
EXAMPLES_DIR = os.path.join(REPO_DIR, 'examples')
PY3 = sys.version_info[0] == 3
WHITELIST = ['grin', 'jetsonconda', 'miniconda', 'newchan']
BLACKLIST = []
WITH_SPACES = {"extra_files", "noconda", "signing", "scripts"}

# .sh installers to also test in interactiv mode
# (require all to a have License = have same interactive input steps)
INTERACTIVE_TESTS = ['miniforge']
# Test runs with even Python version are done in interactive mode, odd in batch mode
INTERACTIVE_TESTING = (sys.version_info.minor % 2) == 0


def _execute(cmd, installer_input=None, **env_vars):
    print(' '.join(cmd))
    t0 = time.time()
    if env_vars:
        env = os.environ.copy()
        env.update(env_vars)
    else:
        env = None
    p = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE if installer_input else None,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        env=env,
    )
    try:
        stdout, stderr = p.communicate(input=installer_input, timeout=420)
        errored = p.returncode != 0
    except subprocess.TimeoutExpired:
        p.kill()
        stdout, stderr = p.communicate()
        print('--- TEST TIMEOUT ---')
        errored = True
    t1 = time.time()
    if errored or "CONDA_VERBOSITY" in env_vars:
        print(f'--- RETURNCODE: {p.returncode} ---')
        if stdout:
            print('--- STDOUT ---')
            print(stdout)
        if stderr:
            print('--- STDERR ---')
            print(stderr)
    print('--- Done in', timedelta(seconds=t1 - t0))
    return errored


def run_examples(keep_artifacts=None, conda_exe=None, debug=False):
    """Run examples bundled with the repository.

    Parameters
    ----------
    keep_artifacts: str, optional=None
        Path where the generated installers will be moved to.
        Will be created if it doesn't exist.

    Returns
    -------
    int
        Number of failed examples
    """
    if sys.platform.startswith("win") and "NSIS_USING_LOG_BUILD" not in os.environ:
        print(
            "! Warning !"
            "  Windows installers are tested with NSIS in silent mode, which does"
            "  not report errors on exit. You should use logging-enabled NSIS builds"
            "  to generate an 'install.log' file this script will search for errors"
            "  after completion."
        )
    example_paths = []
    errored = 0
    if platform.system() != 'Darwin':
        BLACKLIST.append(os.path.join(EXAMPLES_DIR, "osxpkg"))
    if keep_artifacts:
        os.makedirs(keep_artifacts, exist_ok=True)

    whitelist = [os.path.join(EXAMPLES_DIR, p) for p in WHITELIST]
    for fname in os.listdir(EXAMPLES_DIR):
        fpath = os.path.join(EXAMPLES_DIR, fname)
        if os.path.isdir(fpath) and fpath not in whitelist and fpath not in BLACKLIST:
            if os.path.exists(os.path.join(fpath, 'construct.yaml')):
                example_paths.append(fpath)

    # NSIS won't error out when running scripts unless
    # we set this custom environment variable
    os.environ["NSIS_SCRIPTS_RAISE_ERRORS"] = "1"

    parent_output = tempfile.mkdtemp()
    tested_files = set()
    which_errored = {}
    for example_path in sorted(example_paths):
        example_name = Path(example_path).name
        test_with_spaces = example_name in WITH_SPACES
        print(example_name)
        print('-' * len(example_name))
        if (
            sys.platform.startswith("win")
            and conda_exe
            and "micromamba" in os.path.basename(conda_exe).lower()
        ):
            print(
                f"! Skipping {example_name}... Shortcut creation on Windows is "
                "not supported with micromamba."
            )
            continue
        output_dir = tempfile.mkdtemp(prefix=f"{example_name}-", dir=parent_output)
        # resolve path to avoid some issues with TEMPDIR on Windows
        output_dir = str(Path(output_dir).resolve())
        cmd = COV_CMD + ['constructor', '-v', example_path, '--output-dir', output_dir]
        if conda_exe:
            cmd += ['--conda-exe', conda_exe]
        if debug:
            cmd.append("--debug")
        creation_errored = _execute(cmd)
        errored += creation_errored
        for fpath in os.listdir(output_dir):
            ext = fpath.rsplit('.', 1)[-1]
            if fpath in tested_files or ext not in ('sh', 'exe', 'pkg'):
                continue
            tested_files.add(fpath)
            test_suffix = "s p a c e s" if test_with_spaces else None
            env_dir = tempfile.mkdtemp(suffix=test_suffix, dir=output_dir)
            rm_rf(env_dir)
            fpath = os.path.join(output_dir, fpath)
            print('--- Testing', os.path.basename(fpath))
            installer_input = None
            if ext == 'sh':
                if INTERACTIVE_TESTING and example_name in INTERACTIVE_TESTS:
                    cmd = ['/bin/sh', fpath]
                    # Input: Enter, yes to the license, installation folder, no to initialize shells
                    installer_input = f"\nyes\n{env_dir}\nno\n"
                else:
                    cmd = ['/bin/sh', fpath, '-b', '-p', env_dir]
            elif ext == 'pkg':
                if os.environ.get("CI"):
                    # We want to run it in an arbitrary directory, but the options
                    # are limited here... We can only install to $HOME :shrug:
                    # but that will pollute ~, so we only do it if we are running on CI
                    cmd = ['installer', '-pkg', fpath, '-dumplog',
                           '-target', 'CurrentUserHomeDirectory']
                else:
                    # This command only expands the PKG, but does not install
                    cmd = ['pkgutil', '--expand', fpath, env_dir]
            elif ext == 'exe':
                # NSIS manual:
                # > /D sets the default installation directory ($INSTDIR), overriding InstallDir
                # > and InstallDirRegKey. It must be the last parameter used in the command line
                # > and must not contain any quotes, even if the path contains spaces. Only
                # > absolute paths are supported.
                # Since subprocess.Popen WILL escape the spaces with quotes, we need to provide
                # them as separate arguments. We don't care about multiple spaces collapsing into
                # one, since the point is to just have spaces in the installation path -- one
                # would be enough too :)
                # This is why we have this weird .split() thingy down here:
                cmd = ['cmd.exe', '/c', 'start', '/wait', fpath, '/S', *f'/D={env_dir}'.split()]
            env = {"CONDA_VERBOSITY": "3"} if debug else {}
            test_errored = _execute(cmd, installer_input=installer_input, **env)
            # Windows EXEs never throw a non-0 exit code, so we need to check the logs,
            # which are only written if a special NSIS build is used
            win_error_lines = []
            if ext == 'exe' and os.environ.get("NSIS_USING_LOG_BUILD"):
                test_errored = 0
                try:
                    log_is_empty = True
                    with open(os.path.join(env_dir, "install.log"), encoding="utf-16-le") as f:
                        for line in f:
                            log_is_empty = False
                            if ":error:" in line:
                                win_error_lines.append(line)
                                test_errored = 1
                    if log_is_empty:
                        test_errored = 1
                        win_error_lines.append("Logfile was unexpectedly empty!")
                except Exception as exc:
                    test_errored = 1
                    win_error_lines.append(
                        f"Could not read logs! {exc.__class__.__name__}: {exc}\n"
                        "This usually means that the destination folder could not be created.\n"
                        "Possible causes: permissions, non-supported characters, long paths...\n"
                        "Consider setting 'check_path_spaces' and 'check_path_length' to 'False'."
                        )
            for script_prefix in "pre", "post", "test":
                install_location = Path(env_dir)
                if ext == "exe":
                    script_ext = "bat"
                elif ext == "sh":
                    script_ext = "sh"
                elif example_name == "osxpkg":  # we only test one pkg example
                    script_ext = "sh"
                    install_location = Path("~/Library/osx-pkg-test").expanduser()
                else:
                    continue
                if (
                    (Path(example_path) / f"{script_prefix}_install.{script_ext}").exists()
                    and not (install_location / f"{script_prefix}_install_sentinel.txt").exists()
                ):
                    # All pre/post scripts need to write a sentinel file so we can tell they did run
                    test_errored += 1
                    which_errored.setdefault(example_path, []).append(
                        f"Did not find {script_prefix}_install.{script_ext} sentinel!"
                    )
            errored += test_errored
            if test_errored:
                which_errored.setdefault(example_path, []).append(fpath)
                if win_error_lines:
                    print('---  LOGS  ---')
                    for line in win_error_lines:
                        print(line.rstrip())
                if ext == "pkg" and os.environ.get("CI"):
                    # more complete logs are available under /var/log/install.log
                    print('---  LOGS  ---')
                    print("Tip: Debug locally and check the full logs in the Installer UI")
                    print("     or check /var/log/install.log if run from the CLI.")
            elif ext == "exe" and not test_with_spaces:
                # The installer succeeded, test the uninstaller on Windows
                # The un-installers are only tested when testing without spaces, as they hang during
                # testing but work in UI mode.
                uninstaller = next(
                    (p for p in os.listdir(env_dir) if p.startswith("Uninstall-")), None
                )
                if uninstaller:
                    cmd = [
                        'cmd.exe', '/c', 'start', '/wait',
                        os.path.join(env_dir, uninstaller),
                        # We need silent mode + "uninstaller location" (_?=...) so the command can
                        # be waited; otherwise, since the uninstaller copies itself to a different
                        # location so it can be auto-deleted, it returns immediately and it gives
                        # us problems with the tempdir cleanup later
                        f"/S _?={env_dir}"
                    ]
                    test_errored = _execute(cmd)
                    errored += test_errored
                    if test_errored:
                        which_errored.setdefault(example_path, []).append(
                            "Wrong uninstall exit code or timeout."
                        )
                    paths_after_uninstall = os.listdir(env_dir)
                    if len(paths_after_uninstall) > 2:
                        # The debug installer writes to install.log too, which will only
                        # be deleted _after_ a reboot. Finding some files is ok, but more
                        # than two usually means a problem with the uninstaller.
                        # Note this is is not exhaustive, because we are not checking
                        # whether the registry was restored, menu items were deleted, etc.
                        # TODO :)
                        which_errored.setdefault(example_path, []).append(
                            "Uninstaller left too many files behind!\n - "
                            "\n - ".join(paths_after_uninstall)
                        )
                else:
                    which_errored.setdefault(example_path, []).append("Could not find uninstaller!")

            if keep_artifacts:
                dest = os.path.join(keep_artifacts, os.path.basename(fpath))
                if os.path.isfile(dest):
                    os.unlink(dest)
                shutil.move(fpath, keep_artifacts)
        if creation_errored:
            which_errored.setdefault(example_path, []).append("Could not create installer!")
        print()

    print("-------------------------------")
    if errored:
        print('Some examples failed:')
        for installer, reasons in which_errored.items():
            print(f"+ {os.path.basename(installer)}")
            for reason in reasons:
                print(f"---> {reason}")
        print('Assets saved in:', keep_artifacts or parent_output)
    else:
        print('All examples ran successfully!')
        shutil.rmtree(parent_output)
    return errored


def cli():
    p = argparse.ArgumentParser()
    p.add_argument("--keep-artifacts")
    p.add_argument("--conda-exe")
    p.add_argument("--debug", action="store_true", default=False)
    return p.parse_args()


if __name__ == '__main__':
    args = cli()
    if args.conda_exe:
        assert os.path.isfile(args.conda_exe)
    n_errors = run_examples(
        keep_artifacts=args.keep_artifacts,
        conda_exe=args.conda_exe,
        debug=args.debug
    )
    sys.exit(n_errors)
